/*
 *  SAT solution to Kripke model
 *
 *  2015
 *  Valentin Montmirail
 */

#include "Def.h"
#include "Term.h"
#include "Label.h"
#include "Formula.h"
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <string>
#include <algorithm>
#include <vector>
#include <signal.h>
#include <map>
#include <iostream>
#include <cassert>
 #include <unistd.h>


using namespace std;

/** @brief We rename the structure, to avoid typing struct everywhere... */
typedef struct st_kripke st_kripke;

typedef struct st_map st_map;

struct st_map {
    
    int id;

    int value;

    int w_0;

    int w_1;
};

struct st_kripke {
    
    /** @brief How many propositional variables there are in the problem. */
    int num_variables;

    /** @brief How many worlds we use to generate the Kripke model. */
    int num_worlds;

    /** @brief What are the edges from world w_i to world w_j (store as a Matrix [nbWorlds][nbWorlds]). */
    vector<int>** edges;

    /** @brief What are the assignations of propositionals variables for each world (store as a Matrix [nbWorlds][nbVariables]). */
    int** assignations;
};


int nbW_map = 0;
map<string, int> labelWorlds;

unsigned int mappingSize = 0;
int numberVariables = 0;
int maxW = 1;

st_map* mapping;
int* solution;

void alarm_kill(int s) {
  printf("v 0 0 0 0\n" );
  exit(0);
}

/**
 * @brief This function will return the number of edges used in the Kripke model.
 *
 * @param kripke the Kripke structure which store the edges.
 *
 * @return the number of edges which are connecting worlds together.
 */
inline int getNbEdges(st_kripke* kripke) {

  /* at the beginning, we don't know how many edges there is, so 0. */
  int result = 0;
  int i,j;
  
  /* We will check the edge FROM each world w_i */
  for(i = 0 ; i < kripke->num_worlds; i++) {
      /* We will check the edge TO each world w_j */
      result += kripke->edges[i]->size();
  }

  /* And we return at the end how many edges there is in the Kripke structure */
  return result;
}

inline void addEdges(st_kripke* kripke, int w_begin, int w_end) {

  if(std::find(kripke->edges[w_begin]->begin(), kripke->edges[w_begin]->end(), w_end) == kripke->edges[w_begin]->end()) {
    kripke->edges[w_begin]->push_back(w_end);
  }
}

/**
 * @brief This function will initialize a Kripke structure with the good number of variables
 *
 * But the number of worlds may fluctuate (depends of the solving method by the Tableau).
 *
 * @param nbVar the number of propositionals variables.
 * @param nbWorld the original number of worlds (basically the number of modal variables).
 *
 * @return the initialized Kripke structure. 
 */
st_kripke* createStrucKripke(int nbVar, int nbWorld) {

    int i;
    int j;

    /* We need to allocate the place for one Kripke structure. */
    st_kripke* kripke = (st_kripke*)malloc(sizeof(st_kripke)*1);

    kripke->num_variables = nbVar;
    kripke->num_worlds = nbWorld;
    
    /* edges is a matrix [nbWorlds][nbWorlds] so we need to initialize it as an array of arrays */
    kripke->edges = (vector<int>**)malloc(sizeof(vector<int>*)*nbWorld);

    for(i =0 ; i < nbWorld; i++) {
      kripke->edges[i] = new vector<int>();
    }

    /* Then we will need to store each assignations for each world */

    /* So we allocate the place to store an array of integers for every world. */
    kripke->assignations = (int**)malloc(sizeof(int*)*nbWorld);

    /* For each world w_i, we allocate an array with the size nbVar to create a matrix [nbWorld][nbVar] */
    for(i = 0; i < nbWorld; i++) {

      kripke->assignations[i] = (int*)malloc(sizeof(int)*nbVar+1);

      /* Then we initialize each case of theses array with a 0. */
      for(j = 0; j <= nbVar; j++) {
        kripke->assignations[i][j] = 0;
      }
    }

    /* By default, we know that we have one 0, w_0 which is always existing. */
    kripke->num_worlds = nbWorld;

    /* Then we return the perfectly initialized structure. */
    return kripke;
}


void displayStructKripke(st_kripke* model) {

  printf("v %d %d %d %d\n",model->num_variables,model->num_worlds,1,getNbEdges(model));

  for(int i = 0; i < model->num_worlds; i++) {

    int j = 0;
    printf("v ");

    while(j <= model->num_variables) {

      if(model->assignations[i][j] == 0) {
        j++;
        continue;
      }
      
      printf("%d ",model->assignations[i][j]); 
      j++;
    }

    printf("0\n");
  }

  for(int i = 0 ; i < model->num_worlds; i++) {

      unsigned int size = model->edges[i]->size();

      for(int j = 0 ; j < size; j++) {
        printf("v r1 w%d w%d\n",i,model->edges[i][0][j]);
      }
  }

}

int extractSolution(FILE* f) {

  unsigned int size = 65535;
  int i = 0;
  char* line = (char*)malloc(sizeof(char)*size);
  solution = (int*)malloc(sizeof(int)*size);

  while (!feof(f)) {

    fscanf(f,"%s",line);

    if(line[0] == 'U') {
      return 1;
    } 

    if(strcmp(line,"SAT") == 0) {
      printf("s SATISFIABLE\n");
      continue;
    }

    int atoi_line = atoi(line);

    if(i+1 >= size) {
      size *= 2;
      solution = (int*)realloc(solution,sizeof(int)*size);
    }

    solution[i] = atoi_line;
    
    if(solution[i] != 0) numberVariables++;

    i++;

  }

  free(line);

  /* i equals 3, means tautology formula... no need to parse anything. */
  if(i == 3) return -1;

  return 0;
}


unsigned int getIdWorld(const char* label, unsigned int & already) {

  map<string,int>::iterator iter = labelWorlds.find(label);
  if (iter != labelWorlds.end() ) {

    already = 1;
    return iter->second;
  } 
  
  labelWorlds.insert(std::pair<string,int>(label,nbW_map++) );
  already = 0;
  return nbW_map-1;
}

inline void assign(st_kripke* model,int world,int value) {

  model->assignations[world][abs(value)] = value;
}


inline void putValueModel(st_kripke* model) {

  for(unsigned int j = 0 ; j < mappingSize; j++) {
    
      unsigned int abs_sol = abs(solution[j]);

      unsigned world = mapping[j].w_0;

        if(mapping[j].w_1 != -1) {
          world = mapping[j].w_1;
          addEdges(model,mapping[j].w_0,mapping[j].w_1);
        }

        int index = mapping[j].id -1;

        if(solution[index] > 0) {
          assign(model,world,     mapping[j].value);
        }
        else {
          assign(model,world,    -mapping[j].value);
        }
    }
}

void assignWorlds(string & label,int j) {

  mapping[j].w_1 = -1;
  mapping[j].w_0 =  0;
  unsigned int already = 0;
  unsigned int found = 0;

  string label_copy = label;
  const char* clabel = label.c_str();

  if(!strstr(clabel,"_")) {

    mapping[j].w_0 = getIdWorld(clabel,already);

  } else {

      /* We add now the rest of the string, until we arrived to something already knew. */
      do {
          
        found = label.find_last_of("_");
        label = label.substr(0,found);      
        getIdWorld(label.c_str(),already);

      } while(already != 1);
      
      found = label_copy.find_last_of("_");
      label = label_copy.substr(0,found);   
      mapping[j].w_0 = getIdWorld(label.c_str(),already);

      mapping[j].w_1 = getIdWorld(clabel,already);
  }

  if(mapping[j].w_0 >= maxW) maxW = mapping[j].w_0;
  
  if(mapping[j].w_1 >= maxW) maxW = mapping[j].w_1;
}

void assignValueKripke(char* file) {

  int size = 65535;
  int i = 0;
  int j = 0;
  char* world = (char*)malloc(sizeof(char)*size);
  char* varId = (char*)malloc(sizeof(char)*size);
  char* value = (char*)malloc(sizeof(char)*size);

  FILE* f = fopen(file, "r");
  if (f == NULL) {
    fprintf(stderr, "could not open file %s\n", file);
    exit(1);
  } 

  while (!feof(f)) {

    fscanf(f,"%s",varId);
    fscanf(f,"%s",value);

    if(strcmp(value,"false") == 0 || strcmp(value,"true") == 0) {
      continue;
    }

    fscanf(f,"%s",world);

    mapping[j].id = atoi(varId);

    mapping[j].value = atoi(value+1); 

    if(numberVariables <= abs(mapping[j].value)) {
      numberVariables = mapping[j].value;
    }

    string s(world);
    assignWorlds(s,j);

    j++;
  }

  free(world);
  free(varId);
  free(value);
  fclose(f);
  mappingSize = j;
}

int main(int argc, char *argv[]) {

  setbuf(stdout, NULL);
  signal(SIGALRM, alarm_kill);
  signal(SIGINT, alarm_kill);

  int unsat = extractSolution(stdin);
  unsigned int already = 0;

  printf("c Kripke-CNF extraction\n");

  if(unsat == -1) {

    /* Formula tautology. */
    printf("s SATISFIABLE\n");
    printf("c #var #worlds #relations #edges\n");
    printf("d SOLVERCALLS 1\n");
    printf("v 1 1 1 0\n");
    printf("v 0\n");
    return 0;

  } else if(unsat == 1) {
    
    cout << "s UNSATISFIABLE" << endl;
    return 0;
  }

  mapping = (st_map*)malloc(sizeof(st_map)*numberVariables);

  /* Every line starting with a 'c' is a comment. */
  printf("c #var #worlds #relations #edges\n");
  printf("d SOLVERCALLS 1\n");

  /* numberVariables will be modify in assignValueKripke
     to match the p_max, and not the number of SAT variable anymore. */
  numberVariables = 0;

  assignValueKripke(argv[1]);

  st_kripke* model = createStrucKripke(numberVariables,maxW+1);

  putValueModel(model);

  for (const auto &p : labelWorlds) {

      string lab = p.first;

      size_t found = lab.find_last_of("_");

      if(found > 0) {

        string left = lab.substr(0,found);
        string right = lab;

        int w_0 = getIdWorld(left.c_str(),already);
        int w_1 = getIdWorld(right.c_str(),already);
          
        if(w_0 != w_1) addEdges(model,w_0,w_1);
      }
  }

  // for (const auto &p : labelWorlds) {
  //   std::cout << "m[" << p.first << "] = " << p.second << '\n';
  // }

  // exit(0);

  displayStructKripke(model);
  free(mapping);
  free(solution);
  free(model);
  return 0;

}

